unit Unit1;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, FileUtil, Forms, Controls, Graphics, Dialogs, StdCtrls, ExtCtrls, Spin, lNetComponents, lNet,
  PascalModbusTCP;

type
  tran = record
      adr: byte;
      func: byte;
      startReg: word;
      CountReg: word;
  end;

  { TForm1 }

  TForm1 = class(TForm)
      btnConnect: TButton;
      btnDisconn: TButton;
      btnReadH: TButton;
      btnReadI: TButton;
      btnHost: TButton;
      btnDisHost: TButton;
	  btnReadC: TButton;
	  btnReadDI: TButton;
	  btnWrite1C: TButton;
	  btnWrite1H: TButton;
	  btnWriteC: TButton;
	  btnWriteH: TButton;
	  btnMaskWr: TButton;
	  btnRW: TButton;
      editHost: TEdit;
      FNet: TLTCPComponent;
	  Label1: TLabel;
	  Label2: TLabel;
	  Label3: TLabel;
	  Label4: TLabel;
	  Label5: TLabel;
	  Label6: TLabel;
	  Label7: TLabel;
	  Label8: TLabel;
      LServ: TLTCPComponent;
      Memo1: TMemo;
	  spinOrMask: TSpinEdit;
      spinStart: TSpinEdit;
      spinCount: TSpinEdit;
	  spinAndMask: TSpinEdit;
	  spinPort: TSpinEdit;
	  TimerTimeouts: TTimer;
      TimerQuit: TTimer;
      procedure btnConnectClick(Sender: TObject);
      procedure btnDisconnClick(Sender: TObject);
      procedure btnDisHostClick(Sender: TObject);
      procedure btnHostClick(Sender: TObject);
	  procedure btnMaskWrClick(Sender: TObject);
	  procedure btnReadCClick(Sender: TObject);
	  procedure btnReadDIClick(Sender: TObject);
      procedure btnReadHClick(Sender: TObject);
	  procedure btnReadIClick(Sender: TObject);
	  procedure btnRWClick(Sender: TObject);
	  procedure btnWrite1CClick(Sender: TObject);
	  procedure btnWrite1HClick(Sender: TObject);
	  procedure btnWriteCClick(Sender: TObject);
	  procedure btnWriteHClick(Sender: TObject);
      procedure FormClose(Sender: TObject; var CloseAction: TCloseAction);
      procedure FNetConnect(aSocket: TLSocket);
      procedure FNetDisconnect(aSocket: TLSocket);
      procedure FNetError(const msg: string; aSocket: TLSocket);
      procedure FNetReceive(aSocket: TLSocket);
      procedure LServAccept(aSocket: TLSocket);
      procedure LServConnect(aSocket: TLSocket);
      procedure LServDisconnect(aSocket: TLSocket);
      procedure LServError(const msg: string; aSocket: TLSocket);
      procedure LServReceive(aSocket: TLSocket);
      procedure TimerQuitTimer(Sender: TObject);
	  procedure TimerTimeoutsTimer(Sender: TObject);
  private
    { private declarations }
      procedure mbError(Transaction: TModbusTransaction; ErrorCode: byte; aData: rawbytestring);
      procedure mbTimeout(Transaction: TModbusTransaction);
      procedure mbWriteOk(Transaction: TModbusTransaction);
      procedure mbReadBits(Transaction: TModbusTransaction; aData: rawbytestring);
      procedure mbReadWords(Transaction: TModbusTransaction; aData: rawbytestring);
  public
    { public declarations }
  end;

var
  Form1: TForm1;
  bConnected: Boolean = false;
  HoldingReg: array [0..65535] of word;
  InputReg: array [0..65535] of word;
  DInputs: array [0..65535] of Boolean;
  Coils: array [0..65535] of Boolean;
  MaxPDU: word = 1200;
  TransactionID_Control : Boolean = false;
  Transactions: array [0..65535] of tran;
  nextTran : word = 0;
  PlcMem: TPLCMemory;
  PlcMems: TPLCStations;

  mbMaster: TPascalModbasTCPMaster = nil;
  ErrCnt: cardinal = 0;
  RqCnt: cardinal = 0;
  Hosted: Boolean = false;



implementation

{$R *.lfm}

{ TForm1 }

procedure TForm1.FNetConnect(aSocket: TLSocket);
begin
    Memo1.Append('Connected to remote host');
    if mbMaster = nil then
        mbMaster := TPascalModbasTCPMaster.Create;
    // адреса обработчиков
    mbMaster.RequestErrorProcedure := @form1.mbError;
    mbMaster.RequestTimeoutProcedure := @form1.mbTimeout;
    mbMaster.WriteOkProcedure := @form1.mbWriteOk;
    mbMaster.ReadWordsProcedure := @form1.mbReadWords;
    mbMaster.ReadBitsProcedure := @form1.mbReadBits;
    mbMaster.TimeOut := 300;
end;

procedure TForm1.btnConnectClick(Sender: TObject);
begin
    FNet.Connect(editHost.Text, spinPort.Value);
end;

procedure TForm1.btnDisconnClick(Sender: TObject);
begin
    FNet.Disconnect;
end;

procedure TForm1.btnDisHostClick(Sender: TObject);
begin
    LServ.Disconnect;
    if not Hosted then exit;
    Memo1.Append('DisHosted');
    Hosted := false;
end;

procedure TForm1.btnHostClick(Sender: TObject);
var
    i: integer;
begin
    for i := 0 to 255 do PlcMems[i] := @PlcMem;
    if LServ.Connected then exit;
    if Hosted then exit;
    if not LServ.Listen(502) then
    begin
        memo1.Append('Not hosted');
        Hosted := false;
    end
    else
    begin
        Memo1.Append('Hosted');
        Hosted := true;
    end;
end;

procedure TForm1.btnMaskWrClick(Sender: TObject);
var
    s: RawByteString;
    ErrorCode: byte;
    TransID: word;
begin
    if not Assigned(mbMaster) then
        exit;
    s := mbMaster.WriteMaskReg(1, spinStart.Value, spinAndMask.Value, spinOrMask.Value, ErrorCode, TransID);
    Memo1.Append('WriteMaskReg send bytes: ' + inttostr(Length(s)));
    Memo1.Append('Send WriteMaskReg error: ' + inttostr(ErrorCode));
    inc(RqCnt);
    if (s <> '') and (FNet.Connected) then
        FNet.SendMessage(s);
    if ErrorCode <> 0 then
        dec(RqCnt);
    Label4.Caption := inttostr(RqCnt);
end;

procedure TForm1.btnReadCClick(Sender: TObject);
var
    s: RawByteString;
    ErrorCode: byte;
    TransID: word;
begin
    if not Assigned(mbMaster) then
        exit;
    s := mbMaster.ReadCoils(1, spinStart.Value, spinCount.Value, ErrorCode, TransID);
    Memo1.Append('ReadCoils send bytes: ' + inttostr(Length(s)));
    Memo1.Append('Send readC error: ' + inttostr(ErrorCode));
    inc(RqCnt);
    if (s <> '') and (FNet.Connected) then
        FNet.SendMessage(s);
    if ErrorCode <> 0 then
        dec(RqCnt);
    Label4.Caption := inttostr(RqCnt);
end;

procedure TForm1.btnReadDIClick(Sender: TObject);
var
    s: RawByteString;
    ErrorCode: byte;
    TransID: word;
begin
    if not Assigned(mbMaster) then
        exit;
    s := mbMaster.ReadDInputs(1, spinStart.Value, spinCount.Value, ErrorCode, TransID);
    Memo1.Append('btnReadDIClick send bytes: ' + inttostr(Length(s)));
    Memo1.Append('Send readC error: ' + inttostr(ErrorCode));
    inc(RqCnt);
    if (s <> '') and (FNet.Connected) then
        FNet.SendMessage(s);
    if ErrorCode <> 0 then
        dec(RqCnt);
    Label4.Caption := inttostr(RqCnt);
end;

procedure TForm1.btnReadHClick(Sender: TObject);
var
    s: RawByteString;
    ErrorCode: byte;
    TransID: word;
begin
    if not Assigned(mbMaster) then
        exit;
    s := mbMaster.ReadHoldingRegs(1, spinStart.Value, spinCount.Value, ErrorCode, TransID);
    Memo1.Append('ReadHoldingRegs send bytes: ' + inttostr(Length(s)));
    Memo1.Append('Send readH error: ' + inttostr(ErrorCode));
    inc(RqCnt);
    if (s <> '') and (FNet.Connected) then
        FNet.SendMessage(s);
    if ErrorCode <> 0 then
        dec(RqCnt);
    Label4.Caption := inttostr(RqCnt);
end;

procedure TForm1.btnReadIClick(Sender: TObject);
var
    s: RawByteString;
    ErrorCode: byte;
    TransID: word;
begin
    if not Assigned(mbMaster) then
        exit;
    s := mbMaster.ReadInputRegs(1, spinStart.Value, spinCount.Value, ErrorCode, TransID);
    Memo1.Append('ReadInputRegs send bytes: ' + inttostr(Length(s)));
    Memo1.Append('Send readI error: ' + inttostr(ErrorCode));
    inc(RqCnt);
    if (s <> '') and (FNet.Connected) then
        FNet.SendMessage(s);
    if ErrorCode <> 0 then
        dec(RqCnt);
    Label4.Caption := inttostr(RqCnt);
end;

procedure TForm1.btnRWClick(Sender: TObject);
var
    s: RawByteString;
    ErrorCode: byte;
    TransID: word;
    data: RawByteString;
    i: integer;
begin
    if not Assigned(mbMaster) then
        exit;
    data := '';
    for i := spinAndMask.Value to (spinAndMask.Value + spinOrMask.Value - 1) do
        data := data + MB_WordToStrBE(PlcMem.HoldingReg[i]);
    s := mbMaster.ReadWriteRegs(1, spinStart.Value, spinCount.Value, spinAndMask.Value, spinOrMask.Value,
        data, ErrorCode, TransID);
    Memo1.Append('ReadWriteRegs send bytes: ' + inttostr(Length(s)));
    Memo1.Append('Send ReadWriteRegs error: ' + inttostr(ErrorCode));
    inc(RqCnt);
    if (s <> '') and (FNet.Connected) then
        FNet.SendMessage(s);
    if ErrorCode <> 0 then
        dec(RqCnt);
    Label4.Caption := inttostr(RqCnt);
end;

procedure TForm1.btnWrite1CClick(Sender: TObject);
var
    s: RawByteString;
    ErrorCode: byte;
    TransID: word;
begin
    if not Assigned(mbMaster) then
        exit;
    s := mbMaster.WriteOneCoil(1, spinStart.Value, (spinCount.Value <> 0), ErrorCode, TransID);
    Memo1.Append('WriteOneCoil send bytes: ' + inttostr(Length(s)));
    Memo1.Append('Send WriteOneCoil error: ' + inttostr(ErrorCode));
    inc(RqCnt);
    if (s <> '') and (FNet.Connected) then
        FNet.SendMessage(s);
    if ErrorCode <> 0 then
        dec(RqCnt);
    Label4.Caption := inttostr(RqCnt);
end;

procedure TForm1.btnWrite1HClick(Sender: TObject);
var
    s: RawByteString;
    ErrorCode: byte;
    TransID: word;
begin
    if not Assigned(mbMaster) then
        exit;
    s := mbMaster.WriteOneReg(1, spinStart.Value, spinCount.Value, ErrorCode, TransID);
    Memo1.Append('WriteOneReg send bytes: ' + inttostr(Length(s)));
    Memo1.Append('Send WriteOneReg error: ' + inttostr(ErrorCode));
    inc(RqCnt);
    if (s <> '') and (FNet.Connected) then
        FNet.SendMessage(s);
    if ErrorCode <> 0 then
        dec(RqCnt);
    Label4.Caption := inttostr(RqCnt);
end;

procedure TForm1.btnWriteCClick(Sender: TObject);
var
    s: RawByteString;
    ErrorCode: byte;
    TransID: word;
    data: RawByteString;
    i: integer;
begin
    if not Assigned(mbMaster) then
        exit;
    data := '';
    for i := spinStart.Value to (spinStart.Value + spinCount.Value - 1) do
        if PlcMem.Coils[i] then
            data := data + '1'
        else
            data := data + #00;
    s := mbMaster.WriteCoils(1, spinStart.Value, spinCount.Value, data, ErrorCode, TransID);
    Memo1.Append('WriteCoils send bytes: ' + inttostr(Length(s)));
    Memo1.Append('Send WriteCoils error: ' + inttostr(ErrorCode));
    inc(RqCnt);
    if (s <> '') and (FNet.Connected) then
        FNet.SendMessage(s);
    if ErrorCode <> 0 then
        dec(RqCnt);
    Label4.Caption := inttostr(RqCnt);
end;

procedure TForm1.btnWriteHClick(Sender: TObject);
var
    s: RawByteString;
    ErrorCode: byte;
    TransID: word;
    data: RawByteString;
    i: integer;
begin
    if not Assigned(mbMaster) then
        exit;
    data := '';
    for i := spinStart.Value to (spinStart.Value + spinCount.Value - 1) do
        data := data + MB_WordToStrBE(PlcMem.HoldingReg[i]);
    s := mbMaster.WriteCoils(1, spinStart.Value, spinCount.Value, data, ErrorCode, TransID);
    Memo1.Append('WriteCoils send bytes: ' + inttostr(Length(s)));
    Memo1.Append('Send WriteCoils error: ' + inttostr(ErrorCode));
    inc(RqCnt);
    if (s <> '') and (FNet.Connected) then
        FNet.SendMessage(s);
    if ErrorCode <> 0 then
        dec(RqCnt);
    Label4.Caption := inttostr(RqCnt);
end;

procedure TForm1.FormClose(Sender: TObject; var CloseAction: TCloseAction);
begin
  CloseAction := caFree;
  if FNet.Connected then
  begin
      CloseAction := caNone;
      FNet.Disconnect;
      TimerQuit.Enabled := True;
  end;
  if LServ.Connected then
  begin
      CloseAction := caNone;
      LServ.Disconnect;
      TimerQuit.Enabled := True;
  end;
end;

procedure TForm1.FNetDisconnect(aSocket: TLSocket);
begin
    Memo1.Append('disconnect');
    if mbMaster <> nil then
        FreeAndNil(mbMaster);


end;

procedure TForm1.FNetError(const msg: string; aSocket: TLSocket);
begin
    Memo1.Append(msg);
end;

procedure TForm1.FNetReceive(aSocket: TLSocket);
var
    s: AnsiString;
begin
    if mbMaster = nil then exit;
    if aSocket.GetMessage(s) > 0 then
        mbMaster.ReadResponce(s);

end;

procedure TForm1.LServAccept(aSocket: TLSocket);
begin
    Memo1.Append('Accepted from ' + aSocket.PeerAddress + ':' + inttostr(aSocket.PeerPort));
end;

procedure TForm1.LServConnect(aSocket: TLSocket);
begin
    Memo1.Append('Srv connect: ' + aSocket.PeerAddress + ':' + inttostr(aSocket.PeerPort));
end;

procedure TForm1.LServDisconnect(aSocket: TLSocket);
begin
    Memo1.Append('Disconnect: ' + aSocket.PeerAddress + ':' + inttostr(aSocket.PeerPort));
end;

procedure TForm1.LServError(const msg: string; aSocket: TLSocket);
begin
    Memo1.Append('Server error: ' + msg);
end;

procedure TForm1.LServReceive(aSocket: TLSocket);
var
    s, s1: ansistring;
    i: integer = 0;
    ErrorCode: byte;
begin
    s1 := '';
    if aSocket.GetMessage(s) > 0 then
    begin
        { diag info }
        for i := 1 to length(s) do
            s1 := s1 + hexstr(ord(s[i]), 2) + ' ';
        Memo1.Append( '<< ' + s1);

        ErrorCode := 0;
        s1 := ModbusTcPSlaveMessage(PlcMems, ErrorCode, s);
        if ErrorCode <> 0 then
            Memo1.Append( 'Server error: ' + inttostr(ErrorCode));

        if (s1 <> '') then LServ.SendMessage(s1, aSocket);

    end;
end;

procedure TForm1.TimerQuitTimer(Sender: TObject);
begin
    Close;
end;

procedure TForm1.TimerTimeoutsTimer(Sender: TObject);
var
    i: integer;
begin
    if Assigned(mbMaster) then
        mbMaster.CheckTimeouts;
    //for i := 0  to 65535 do
    //    PlcMem.InputReg[i] := PlcMem.HoldingReg[i];
    //for i := 0  to 65535 do
    //    PlcMem.DInputs[i] := PlcMem.Coils[i];
end;

procedure TForm1.mbError(Transaction: TModbusTransaction; ErrorCode: byte; aData: rawbytestring);
var
	i: Integer;
    s1: rawbytestring;
begin
    inc(ErrCnt);
    Label4.Caption := inttostr(ErrCnt);
    Memo1.Append('Request error: ' + inttostr(ErrorCode));
    { diag info }
    for i := 1 to length(aData) do
        s1 := s1 + hexstr(ord(aData[i]), 2) + ' ';
    Memo1.Append( '<< ' + s1);

end;

procedure TForm1.mbTimeout(Transaction: TModbusTransaction);
begin
    inc(ErrCnt);
    Label4.Caption := inttostr(ErrCnt);
    Memo1.Append('Timeout ');
end;

procedure TForm1.mbWriteOk(Transaction: TModbusTransaction);
begin
    Memo1.Append('Write ok');
end;

procedure TForm1.mbReadBits(Transaction: TModbusTransaction; aData: rawbytestring);
var
    i: word;
begin
    if Length(aData) <> Transaction.CountReg then
    begin
        Memo1.Append('mbReadBits Error');
        exit;
    end;
    if Transaction.Func = 01 then
    begin
        for i := 0 to Transaction.CountReg - 1 do
            if aData[i+1] <> #00 then
                PlcMem.Coils[Transaction.StartReg + i] := true
            else
                PlcMem.Coils[Transaction.StartReg + i] := false;
    end;
    if Transaction.Func = 02 then
    begin
        for i := 0 to Transaction.CountReg - 1 do
            if aData[i+1] <> #00 then
                PlcMem.DInputs[Transaction.StartReg + i] := true
            else
                PlcMem.DInputs[Transaction.StartReg + i] := false;
    end;
end;

procedure TForm1.mbReadWords(Transaction: TModbusTransaction; aData: rawbytestring);
var
	i: word;
begin
    if Length(aData) <> (Transaction.CountReg * 2) then
    begin
        Memo1.Append('mbReadWords Error');
        exit;
    end;
    if Transaction.Func = 03 then
    begin
        Memo1.Append('Get func 3 data');
        for i := 0 to Transaction.CountReg - 1 do
            PlcMem.HoldingReg[Transaction.StartReg + i] := MB_StrToWord(aData[i*2+1], aData[i*2+2]);
    end;
    if Transaction.Func = 04 then
    begin
        Memo1.Append('Get func 4 data');
        for i := 0 to Transaction.CountReg - 1 do
            PlcMem.InputReg[Transaction.StartReg + i] := MB_StrToWord(aData[i*2+1], aData[i*2+2]);
    end;
end;

end.

