{
Copyright (c) 2016, Yuri Lychakov
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL YURI LYCHAKOV BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
}
unit Pascal_Modbus_RTU;

{ ENG: Comments on Russian! Translation will be in future versions.
 At the moment, I suggest to use automatic translation.}

{$mode objfpc}{$H+}
{ TODO :     }
{ TODO :       ,   20-23 }
{ TODO :         }
interface

uses
    Classes, SysUtils, synaser, SynaUtil, crc16_stream, Windows;

type

    //  -     
    TProcedureReadModbus = procedure(Station: byte; FunctionCode: byte;
        StartAddress: word; CountSended: word; size_mb: cardinal; DataStream: TMemoryStream;
        NotMatch: boolean = False) of object;
    //  -    
    TResponceError = procedure(Station: byte; ErrorCode, ExceptionCode: byte) of object;
    //  -    
    TListenError = procedure(ErrorCode: byte; DataStream: TmemoryStream) of object;
    //     
    TDiagStr = procedure(DiagStr: string) of object;


    { TSendThread }

    //        
    TSendThread = class(TThread)
    private
    protected
        procedure Execute; override;
    public
        SendStream: TMemoryStream; //  ,   ""   ()
        serial: TBlockSerial;   // ,    
        RTS: boolean;   //    RTS   
        RTS_On: word;   // -    RTS   
        RTS_Off: word;  // -       RTS
        constructor Create(CreateSuspended: boolean);
    end;

    DCB = record
          DCBlength : DWORD;
          BaudRate : DWORD;
          flags : DWORD;
          wReserved : WORD;
          XonLim : WORD;
          XoffLim : WORD;
          ByteSize : BYTE;
          Parity : BYTE;
          StopBits : BYTE;
          XonChar : char;
          XoffChar : char;
          ErrorChar : char;
          EofChar : char;
          EvtChar : char;
          wReserved1 : WORD;
       end;

    { TPascal_Modbus_RTU }

    TPascal_Modbus_RTU = class(TObject)
    private
        FConnected: boolean;        //   
        FDiagStr: TDiagStr;         //   .  ()
        FIsMaster: boolean;         //  
        FLastErrorDesc: string;     //    ( )
        FListenError: TListenError; //     
        FListenErrorCode: cardinal; //     
        FMaxSizePDU: cardinal;      //     ModBus
        FObjectTag: integer;
        FResponceBreakTime: cardinal;   //  ( )    " "
        FResponceError: TResponceError; //    
        FResponceReadProcedure: TProcedureReadModbus;  //    
        FRTS_Off_Delay: word;       // -       RTS
        FRTS_On_Delay: word;        // -    RTS   
        FRTS_use: boolean;          //   RTS   
        FSendErrorMessage: boolean; //  "    "
        FSilenceCount: cardinal;    //   ""  , 
        FTimeBeforeSend: cardinal;  //   ""     , 
        FTimeOutResponce: cardinal; //     , 
        FReadyForSend: boolean;     //  "   "

        FirstStart: boolean;        //  " "
        FTimeToSleepRx: cardinal;   //        (?)
        TimeToSleepTx: cardinal;    //        
        InputStream: TMemoryStream; //     
        OutputStream: TMemoryStream;  //     
        InputMessageStream: TMemoryStream;  // ???  -  (?)
        SendStream: TMemoryStream;  //      
        SendedSize: word;           //   
        SendedFunction: byte;       //  ,  
        SendedStation: byte;        //  ,   
        SendedAddress: word;        //    
        Serial: TBlockSerial;       // COM-
        BaudRate: cardinal;         //  
        MessageFromMaster: boolean; // ---  ,     
        SendThread: TSendThread;    //   
        WaitingResponce: boolean;   //  " "
        stime: tdatetime;           //   

        procedure SendFinished(Sender: TObject);
        procedure SetDiagStr(AValue: TDiagStr);

        procedure SetIsMaster(AValue: boolean);
        procedure SetListenError(AValue: TListenError);
        procedure SetMaxSizePDU(AValue: cardinal);
        procedure SetObjectTag(AValue: integer);
        procedure SetResponceBreakTime(AValue: cardinal);
        procedure SetResponceError(AValue: TResponceError);
        procedure SetResponceReadProcedure(AValue: TProcedureReadModbus);
        procedure SetRTS_Off_Delay(AValue: word);
        procedure SetRTS_On_Delay(AValue: word);
        procedure SetRTS_use(AValue: boolean);
        procedure SetSendErrorMessage(AValue: boolean);
        procedure SetTimeOutResponce(AValue: cardinal);
        procedure SetTimeToSleepRx(AValue: cardinal);
    protected

    public
        //   
        property LastErrorDesc: string read FLastErrorDesc ;
        //     
        property ListenErrorCode: cardinal read FListenErrorCode;
        //   ""  , 
        property SilenceCount: cardinal read FSilenceCount;
        //   ""     , 
        property TimeBeforeSend: cardinal read FTimeBeforeSend;
        //   
        property Connected: boolean read FConnected;
        //  "   "
        property ReadyForSend: boolean read FReadyForSend;

        constructor Create;
        destructor Free;
        //     ( ... )
        function Read(Station, FunctionCode: byte; StartAddress, Count: word): cardinal;
        //     ( ... )
        function Write(Station, FunctionCode: byte; StartAddress, Count: word;
            WriteStream: Tmemorystream): cardinal;
        //   COM-
        function ConnectCOM(ComPort: string): integer;
        //   
        procedure DisconnectCom;
        //   COM- ( )
        procedure ConfigCOM(baud, bits: integer; parity: char; stop: integer;
            softflow, hardflow: boolean);
        procedure ListenLine;

        //     , 
        property TimeOutResponce: cardinal read FTimeOutResponce write SetTimeOutResponce;
        //  "    "
        property SendErrorMessage: boolean read FSendErrorMessage write SetSendErrorMessage;
        //    
        property ResponceReadProcedure: TProcedureReadModbus
            read FResponceReadProcedure write SetResponceReadProcedure;
        //    
        property ResponceError: TResponceError read FResponceError write SetResponceError;
        //    
        property ListenError: TListenError read FListenError write SetListenError;
        { TODO :  ObjectTag -   }
        property ObjectTag: integer read FObjectTag write SetObjectTag;
        //  -  ( ?)
        property IsMaster: boolean read FIsMaster write SetIsMaster;
        //     ModBus
        property MaxSizePDU: cardinal read FMaxSizePDU write SetMaxSizePDU;
        //      
        property TimeToSleepRx: cardinal read FTimeToSleepRx write SetTimeToSleepRx;
        //  ( )    " "
        property ResponceBreakTime: cardinal read FResponceBreakTime write SetResponceBreakTime;
        //   RTS   
        property RTS_use: boolean read FRTS_use write SetRTS_use;
        // -    RTS   
        property RTS_On_Delay: word read FRTS_On_Delay write SetRTS_On_Delay;
        // -       RTS
        property RTS_Off_Delay: word read FRTS_Off_Delay write SetRTS_Off_Delay;
        //   
        property DiagStr: TDiagStr read FDiagStr write SetDiagStr;
    end;

    { TTimerThread }

    TTimerThread = class(TThread)
    private
        procedure Timer;
    protected
        procedure Execute; override;
    public
        StopTimer: Boolean;
        mbrtu: TPascal_Modbus_RTU;
        constructor Create(CreateSuspended: boolean);
    end;

implementation

{ TPascal_Modbus_RTU }
var
    TimerThread: TTimerThread;

procedure TPascal_Modbus_RTU.SetTimeOutResponce(AValue: cardinal);
begin
    if FTimeOutResponce = AValue then
        Exit;
    FTimeOutResponce := AValue;
end;

procedure TPascal_Modbus_RTU.SetTimeToSleepRx(AValue: cardinal);
begin
    if FTimeToSleepRx = AValue then
        Exit;
    FTimeToSleepRx := AValue;
end;

procedure TPascal_Modbus_RTU.SetSendErrorMessage(AValue: boolean);
begin
    if FSendErrorMessage = AValue then
        Exit;
    FSendErrorMessage := AValue;
end;

procedure TPascal_Modbus_RTU.SetResponceReadProcedure(AValue: TProcedureReadModbus);
begin
    if FResponceReadProcedure = AValue then
        Exit;
    FResponceReadProcedure := AValue;
end;

procedure TPascal_Modbus_RTU.SetRTS_Off_Delay(AValue: word);
begin
    if FRTS_Off_Delay = AValue then
        Exit;
    FRTS_Off_Delay := AValue;
end;

procedure TPascal_Modbus_RTU.SetRTS_On_Delay(AValue: word);
begin
    if FRTS_On_Delay = AValue then
        Exit;
    FRTS_On_Delay := AValue;
end;

procedure TPascal_Modbus_RTU.SetRTS_use(AValue: boolean);
begin
    if FRTS_use = AValue then
        Exit;
    FRTS_use := AValue;
end;

procedure TPascal_Modbus_RTU.SetResponceError(AValue: TResponceError);
begin
    if FResponceError = AValue then
        Exit;
    FResponceError := AValue;
end;

procedure TPascal_Modbus_RTU.SetObjectTag(AValue: integer);
begin
    if FObjectTag = AValue then
        Exit;
    FObjectTag := AValue;
end;

procedure TPascal_Modbus_RTU.SetResponceBreakTime(AValue: cardinal);
begin
    if FResponceBreakTime = AValue then
        Exit;
    FResponceBreakTime := AValue;
end;

procedure TPascal_Modbus_RTU.SetDiagStr(AValue: TDiagStr);
begin
    if FDiagStr = AValue then Exit;
    FDiagStr := AValue;
end;

procedure TPascal_Modbus_RTU.SetIsMaster(AValue: boolean);
begin
    if FConnected then
        exit;
    if FIsMaster = AValue then
        Exit;
    FIsMaster := AValue;
end;

procedure TPascal_Modbus_RTU.SetListenError(AValue: TListenError);
begin
    if FListenError = AValue then Exit;
    FListenError := AValue;
end;

procedure TPascal_Modbus_RTU.SetMaxSizePDU(AValue: cardinal);
begin
    if FMaxSizePDU = AValue then
        Exit;
    FMaxSizePDU := AValue;
end;


constructor TPascal_Modbus_RTU.Create;
begin
    TimerThread := TTimerThread.Create(True);
    TimerThread.mbrtu := self;
    if Assigned(TimerThread.FatalException) then
        raise TimerThread.FatalException;
    TimerThread.Start;

    FLastErrorDesc := '';
    InputStream := TMemoryStream.Create;
    FObjectTag := 0;
    FResponceError := nil;
    FResponceReadProcedure := nil;
    FSendErrorMessage := False;
    FSilenceCount := 0;
    FTimeOutResponce := 400;
    OutputStream := nil;
    SendStream := nil;
    WaitingResponce := False;
    InputMessageStream := TMemoryStream.Create;
    SendedSize := 0;
    SendedFunction := 0;
    SendedStation := 0;
    SendedAddress := 0;
    FTimeBeforeSend := 0;
    FListenErrorCode := 0;
    FRTS_On_Delay := 100;
    FRTS_Off_Delay := 100;
    BaudRate := 9600;
    FRTS_use := False;
    FMaxSizePDU := 253;
    TimeToSleepTx := 1;
    FTimeToSleepRx := 20;
    FResponceBreakTime := 20;
    FConnected := False;
    FReadyForSend := False;
    Serial := TBlockSerial.Create;
    Serial.RaiseExcept := False;
    FIsMaster := True;
    MessageFromMaster := False;
end;

destructor TPascal_Modbus_RTU.Free;
begin
    if Assigned(Serial) then
    begin
        Serial.CloseSocket; //  
        FreeAndNil(Serial);
    end;
    FreeAndNil(InputStream);
    FreeAndNil(InputMessageStream);
    TimerThread.Terminate;
    FreeAndNil(TimerThread);
    inherited Free;
end;

procedure TPascal_Modbus_RTU.SendFinished(Sender: TObject);
begin
    SendThread := nil;
    FReadyForSend := True;
    WaitingResponce := True;
    TimeToSleepTx := {(SendStream.Size * 10000 div BaudRate) +} FTimeOutResponce;

end;

function TPascal_Modbus_RTU.Read(Station, FunctionCode: byte; StartAddress, Count: word): cardinal;
begin
    Result := 1;
    if FTimeBeforeSend <> 0 then
        exit;
    Result := 2;
    if Serial.SendingData > 0 then
        exit;
    Result := 3;
    if not Serial.CanWrite(0) then
        exit;
    Result := 4;
    if not FReadyForSend then
        exit;
    Result := 5;
    if WaitingResponce then
        exit;
    Result := 6;
    if not FunctionCode in [1, 2, 3, 4] then
        exit;
    Result := 7;
    if ((Count > ((FMaxSizePDU - 3) * 8)) and (FunctionCode in [1, 2])) or
        ((Count > ((FMaxSizePDU - 3) div 2)) and (FunctionCode in [3, 4])) then
        exit;
    Result := 8;
    if (StartAddress + Count) > 65535 then
        exit;

    //TimerThread.StopTimer := true;

    SendStream := TMemoryStream.Create;
    FReadyForSend := False;
    SendedSize := Count;
    SendedAddress := StartAddress;
    SendedFunction := FunctionCode;
    SendedStation := Station;

    SendStream.WriteByte(Station);
    SendStream.WriteByte(FunctionCode);
    SendStream.WriteByte(byte(StartAddress shr 8));
    SendStream.WriteByte(byte(StartAddress and $FF));
    SendStream.WriteByte(byte(Count shr 8));
    SendStream.WriteByte(byte(Count and $FF));
    add_crc16(SendStream);
    SendStream.Position := 0;

    //TimeToSleepTx := {(SendStream.Size * 10000 div BaudRate) +} FTimeOutResponce;
    try
        //    
        //InputStream.SetSize(InputStream.Size + waiting);
        //      
        serial.RecvStreamSize(InputStream, 1, serial.WaitingDataEx);
        Serial.Purge;
        //    -  
        //InputStream.Position := 0;
    except
        FListenErrorCode := 3;
        if Assigned(FListenError) then FListenError(0, nil);
    end;
    InputStream.Position := 0;
    InputStream.Clear;

    SendThread := TSendThread.Create(True);
    SendThread.SendStream := SendStream;
    SendThread.serial := Serial;
    SendThread.OnTerminate := @SendFinished;
    SendThread.RTS := FRTS_use;
    SendThread.RTS_On := FRTS_On_Delay;
    SendThread.RTS_Off := FRTS_Off_Delay;
    if Assigned(SendThread.FatalException) then
        raise SendThread.FatalException;
    SendThread.Start;

    //TimerThread.StopTimer := false;

    Result := 0;
end;

function TPascal_Modbus_RTU.Write(Station, FunctionCode: byte; StartAddress, Count: word;
    WriteStream: Tmemorystream): cardinal;
var
    bt: byte;
begin
    Result := 1;
    if FTimeBeforeSend <> 0 then
        exit;
    Result := 2;
    if Serial.SendingData > 0 then
        exit;
    Result := 3;
    if not Serial.CanWrite(0) then
        exit;
    Result := 4;
    if not FReadyForSend then
        exit;
    Result := 5;
    if WaitingResponce then
        exit;
    Result := 6;
    if not FunctionCode in [15, 16] then
        exit;
    Result := 7;
    if ((Count > ((FMaxSizePDU - 7) * 8)) and (FunctionCode = 15)) or
        ((Count > ((FMaxSizePDU - 3) div 2)) and (FunctionCode = 16)) then
        exit;
    Result := 8;
    if ((WriteStream.Size > (FMaxSizePDU - 7)) or (WriteStream.Size = 0)) then
        exit;
    Result := 9;
    if (FunctionCode = 15) and (WriteStream.Size < (Count div 8)) or
        (WriteStream.Size < (Count div 8 + 1)) then
        exit;
    Result := 10;
    if (FunctionCode = 16) and (WriteStream.Size <> (Count * 2)) then
        exit;
    Result := 11;
    if (StartAddress + Count) > 65535 then
        exit;
    WriteStream.Position := 0;
    SendStream := TMemoryStream.Create;
    FReadyForSend := False;
    SendedSize := Count;
    SendedAddress := StartAddress;
    SendedFunction := FunctionCode;
    SendedStation := Station;

    SendStream.WriteByte(Station);
    SendStream.WriteByte(FunctionCode);
    SendStream.WriteByte(byte(StartAddress shr 8));
    SendStream.WriteByte(byte(StartAddress and $FF));
    SendStream.WriteByte(byte(Count shr 8));
    SendStream.WriteByte(byte(Count and $FF));
    bt := WriteStream.Size;
    SendStream.WriteByte(bt);
    for bt := 1 to WriteStream.Size do
        SendStream.WriteByte(WriteStream.ReadByte);

    add_crc16(SendStream);
    SendStream.Position := 0;

    //TimeToSleepTx := {(SendStream.Size * 10000 div BaudRate) +} FTimeOutResponce;
    try
        //    
        //InputStream.SetSize(InputStream.Size + waiting);
        //      
        serial.RecvStreamSize(InputStream, 1, serial.WaitingDataEx);
        //    -  
        //InputStream.Position := 0;
    except
        FListenErrorCode := 3;
        if Assigned(FListenError) then FListenError(0, nil);
    end;
    InputStream.Position := 0;
    InputStream.Clear;

    //WaitingResponce := True;
    SendThread := TSendThread.Create(True);
    SendThread.SendStream := SendStream;
    SendThread.serial := Serial;
    SendThread.OnTerminate := @SendFinished;
    SendThread.RTS := FRTS_use;
    SendThread.RTS_On := FRTS_On_Delay;
    SendThread.RTS_Off := FRTS_Off_Delay;
    if Assigned(SendThread.FatalException) then
        raise SendThread.FatalException;
    SendThread.Start;

    FreeAndNil(WriteStream);
    Result := 0;
end;

function TPascal_Modbus_RTU.ConnectCOM(ComPort: string): integer;
begin
    Result := 1;
    try
        Serial.Connect(ComPort);
    except
        exit;
    end;
    Result := Serial.LastError;
    FLastErrorDesc := Serial.LastErrorDesc;
    if Serial.LastError <> 0 then
        exit;
    FConnected := True;
    FReadyForSend := True;
    FirstStart := True;
    FTimeBeforeSend := 5;
end;

procedure TPascal_Modbus_RTU.DisconnectCom;
begin
    if not FConnected then
        exit;
    Serial.CloseSocket;
    FConnected := False;
    FReadyForSend := False;
    TimerThread.Terminate;
end;

procedure TPascal_Modbus_RTU.ConfigCOM(baud, bits: integer; parity: char; stop: integer;
    softflow, hardflow: boolean);

begin
    if FConnected then
        Serial.Config(baud, bits,parity, stop, softflow, hardflow);
    BaudRate := baud;
    sleep(100);

end;

procedure TPascal_Modbus_RTU.ListenLine;
var
    waiting: integer;  //      
    StId, MbFunc, sz: byte;
    i: integer;
begin
    //      .    -  
    if FTimeBeforeSend <> 0 then
        Dec(FTimeBeforeSend);

    //    (   com-),  ,    
    FListenErrorCode := 1;
    if not FConnected then
    begin
        exit;
    end;
    //         
    waiting := Serial.WaitingDataEx;

    if waiting > 0 then
        FTimeBeforeSend := FTimeToSleepRx {+ waiting * 10000 div BaudRate };

    FListenErrorCode := 2;
    //        ,     
    if FirstStart then
        if waiting > 0 then
        begin
            exit
        end
        else
        begin
            //  ,   -   ()
            FirstStart := False;
            //      ,   _.
            if IsMaster then
                MessageFromMaster := False
            else
                MessageFromMaster := True;
        end;
    FListenErrorCode := 0;
    //     
    if waiting = 0 then
    begin
        if FSilenceCount = 0 then stime := now;
        //  " "
        Inc(FSilenceCount);
        //      
        if FSilenceCount > 4001001001 then
            FSilenceCount := 4001001001;

        //    
        if (InputStream.Size > 0) and (FSilenceCount >= FResponceBreakTime) then
        begin
            InputStream.Position := 0;
            OutputStream := TMemoryStream.Create;
            InputStream.SaveToStream(OutputStream);
            InputStream.Position := 0;
            OutputStream.Position := 0;
            InputStream.Clear;

            if Assigned(FListenError) then FListenError(2, OutputStream);
            //if Assigned(DiagStr) then DiagStr(inttostr(FSilenceCount) + '  t: ' + formatdatetime('ss.zzz', now - stime) );

            WaitingResponce := False;
            { TODO : Generate error BreakResponce }
        end;


        if WaitingResponce and (TimeToSleepTx = 0) then
        begin
            WaitingResponce := False;
            if Assigned(FListenError) then FListenError(1, nil);
        end;
        if WaitingResponce and (TimeToSleepTx > 0) then
            Dec(TimeToSleepTx);
        exit;
    end;

    //    , .. com- - 
    //if Assigned(FDiagStr) then FDiagStr(IntToStr(FSilenceCount) + ' :  ' + IntToStr(waiting));
    try
        //    
        //InputStream.SetSize(InputStream.Size + waiting);
        //      
        serial.RecvStreamSize(InputStream, 0, waiting);
    except
        FListenErrorCode := 3;
        if Assigned(FListenError) then FListenError(0, nil);
    end;

    //  -  -   
    FSilenceCount := 0;
    TimeToSleepTx := FTimeOutResponce;

    //    -          
    if FReadyForSend = false then
    begin
      InputStream.Position := 0;
      InputStream.Clear;
    end;

    if InputStream.Size < 4 then
    begin
        exit;
    end;
    if not check_crc16(InputStream, InputStream.Size - 2) then
    begin
        if InputStream.Size > FMaxSizePDU + 3 then
        begin
            InputStream.Position := 0;
            OutputStream := TMemoryStream.Create;
            InputStream.SaveToStream(OutputStream);
            InputStream.Position := 0;
            OutputStream.Position := 0;
            InputStream.Clear;

            FListenErrorCode := 4;
             WaitingResponce := False;
        if Assigned(FListenError) then FListenError(4, OutputStream);
        end;
        exit;
    end;

    InputStream.Position := 0;
    stid := InputStream.ReadByte;
    MbFunc := InputStream.ReadByte;
    if (MbFunc in [3, 4]) then
        sz := InputStream.ReadByte
    else
        sz := InputStream.Size - 4; //    Hilding- -    -

    if (MbFunc in [3, 4]) and ((sz mod 2) = 1) then
    begin
        FListenErrorCode := 5;
        InputStream.Position := 0;
        InputStream.Clear;
        FListenErrorCode := 0;
        WaitingResponce := False;
        if Assigned(FListenError) then FListenError(3, nil);
        exit;
    end;

    if (MbFunc > 127) then
    begin
        FListenErrorCode := 6;
        InputStream.Position := 0;
        InputStream.Clear;
        FListenErrorCode := 0;
        WaitingResponce := False;
        if Assigned(FListenError) then FListenError(MbFunc, nil);
        exit;
    end;

    OutputStream := TMemoryStream.Create;
    for i := 0 to sz - 1 do
    begin
        OutputStream.WriteByte(InputStream.ReadByte);
    end;
    InputStream.Position := 0;
    InputStream.Clear;
    FListenErrorCode := 0;
    WaitingResponce := False;

    if Assigned(FResponceReadProcedure) then
    begin
        FResponceReadProcedure(stid, MbFunc, SendedAddress, SendedSize, sz, OutputStream);
    end;
end;


{ TSendThread }

procedure TSendThread.Execute;
begin
    if RTS then
    begin
        serial.RTS := True;
        sleep(RTS_On);
    end;
    Serial.SendStreamRaw(SendStream);
    FreeAndNil(SendStream);
    if RTS then
    begin
        sleep(RTS_Off);
        serial.RTS := False;
    end;
    Terminate;
end;

constructor TSendThread.Create(CreateSuspended: boolean);
begin
    FreeOnTerminate := True;
    inherited Create(CreateSuspended);
end;

{ TTimerThread }

procedure TTimerThread.timer;
begin
    if not Assigned(mbrtu) then
    begin
        Terminate;
        exit;
    end;
    if mbrtu.Connected then
        mbrtu.ListenLine;

end;

procedure TTimerThread.Execute;
begin
    while (not Terminated) do
    begin
        sleep(1);
            Synchronize(@timer);
    end;
end;

constructor TTimerThread.Create(CreateSuspended: boolean);
begin
    FreeOnTerminate := True;
    StopTimer := false;
    inherited Create(CreateSuspended);
end;

end.
