{
Copyright (c) 2016, Yuri Lychakov
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL YURI LYCHAKOV BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
}
unit crc16_stream;

{$mode objfpc}{$H+}

interface



uses
    Classes, SysUtils;

function crc16_stream(AStream: tstream): word;
function check_crc16(AStream: tstream; size: cardinal): Boolean;
procedure add_crc16(AStream: tstream);

implementation

function crc16_stream(AStream: tstream): word;
var
    crc: word;
    i: integer;
    j: byte;
    b: byte;
begin
    AStream.Position := 0;
    crc := $FFFF;
    for i := 0 to AStream.Size - 1 do
    begin
        b := AStream.ReadByte;
        crc := crc xor b;
        for j := 0 to 7 do
        begin
            if (crc and 1) = 1 then
                crc := (crc shr 1) xor $a001
            else
                crc := crc shr 1;
        end;
    end;
    Result := crc;
end;

function check_crc16(AStream: tstream; size: cardinal): Boolean;
var
    crc: word;
    i: integer;
    j: byte;
    b: byte;
    crc_InStream: word;
begin
    Result:= false;
    if AStream.Size < (size + 2) then exit;
    AStream.Position := 0;
    crc := $FFFF;
    for i := 0 to size - 1 do
    begin
        b := AStream.ReadByte;
        crc := crc xor b;
        for j := 0 to 7 do
        begin
            if (crc and 1) = 1 then
                crc := (crc shr 1) xor $a001
            else
                crc := crc shr 1;
        end;
    end;
    crc_InStream := AStream.ReadByte or (AStream.ReadByte shl 8);
    Result := (crc_InStream = crc);
end;

procedure add_crc16(AStream: tstream);
var
    crc: word;
    i: integer;
    j: byte;
    b: byte;
begin
    AStream.Position := 0;
    crc := $FFFF;
    for i := 0 to AStream.Size - 1 do
    begin
        b := AStream.ReadByte;
        crc := crc xor b;
        for j := 0 to 7 do
        begin
            if (crc and 1) = 1 then
                crc := (crc shr 1) xor $a001
            else
                crc := crc shr 1;
        end;
    end;
    AStream.WriteByte(crc);
    AStream.WriteByte(crc shr 8);
end;


end.

