{
Copyright (c) 2016, Yuri Lychakov
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL YURI LYCHAKOV BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
}

unit PascalModbusTCP;

{ ENG: Comments on Russian! Translation will be in future versions.
 At the moment, I suggest to use automatic translation.}

{$mode objfpc}{$H+}

interface

uses
    Classes, SysUtils, dateutils;

type

    TPLCMemory = record
        DInputs: array [0..65535] of boolean;
        Coils: array [0..65535] of boolean;
        InputReg: array [0..65535] of word;
        HoldingReg: array [0..65535] of word;
    end;

    TPLCStations = array [0..255] of ^TPLCMemory;

    TModbusTransaction = record
        Adress: byte;
        Func: byte;
        StartReg: word;
        CountReg: word;
        WriteReg: word;
        WriteCount: word;
        StartTime: word;
    end;

    //  
    TModbusTCPRequestTimeout = procedure(Transaction: TModbusTransaction) of object;
    //  
    TModbusTCPRequestError = procedure(Transaction: TModbusTransaction; ErrorCode: byte; aData: rawbytestring) of object;
    //   
    TModbusTCPWriteOk = procedure(Transaction: TModbusTransaction) of object;
    //    (01, 02 )
    TModbusTCPReadBits = procedure(Transaction: TModbusTransaction; aData: rawbytestring) of object;
    //    (03, 04, 23 )
    TModbusTCPReadWords = procedure(Transaction: TModbusTransaction; aData: rawbytestring) of object;

    { TPascalModbasTCPMaster }

    TPascalModbasTCPMaster = class(TObject)
    private
        //    (   )
        FActiveRequests: word;
        //     (   )
        FMaxActiveRequests: word;
        //  
        FRequestTimeoutProcedure: TModbusTCPRequestTimeout;
        //  
        FRequestErrorProcedure: TModbusTCPRequestError;
        FSwapBytes: boolean;
        //   
        FWriteOkProcedure: TModbusTCPWriteOk;
        //    (01, 02 )
        FReadBitsProcedure: TModbusTCPReadBits;
        //    (03, 04, 23 )
        FReadWordsProcedure: TModbusTCPReadWords;


        FTimeout: word;

        NextTransID: word;
        TransMap1: array [0..31,0..31,0..63] of boolean;
        TransMap2: array [0..31,0..31] of boolean;
        TransMap3: array [0..31] of boolean;

        procedure SetMaxActiveRequests(AValue: word);
        procedure SetReadBitsProcedure(AValue: TModbusTCPReadBits);
        procedure SetReadWordsProcedure(AValue: TModbusTCPReadWords);
        procedure SetRequestErrorProcedure(AValue: TModbusTCPRequestError);
        procedure SetRequestTimeoutProcedure(AValue: TModbusTCPRequestTimeout);
        procedure SetSwapBytes(AValue: boolean);
        procedure SetTimeOut(AValue: word);
        procedure SetWriteOkProcedure(AValue: TModbusTCPWriteOk);

        procedure GetNextTransID;
        procedure DeactivateTransaction(const RTransID: word);

    public
        Transactions: array [0..65535] of TModbusTransaction;

        //    (   )
        property ActiveRequests: word read FActiveRequests;
        //     (   )
        property MaxActiveRequests: word read FMaxActiveRequests write SetMaxActiveRequests;

        //  (    )
        property TimeOut: word read FTimeOut write SetTimeOut;

        //  
        property RequestTimeoutProcedure: TModbusTCPRequestTimeout read FRequestTimeoutProcedure
            write SetRequestTimeoutProcedure;
        //  
        property RequestErrorProcedure: TModbusTCPRequestError read FRequestErrorProcedure write SetRequestErrorProcedure;
        //   
        property WriteOkProcedure: TModbusTCPWriteOk read FWriteOkProcedure write SetWriteOkProcedure;
        //    (01, 02 )
        property ReadBitsProcedure: TModbusTCPReadBits read FReadBitsProcedure write SetReadBitsProcedure;
        //    (03, 04, 23 )
        property ReadWordsProcedure: TModbusTCPReadWords read FReadWordsProcedure write SetReadWordsProcedure;
        property SwapBytes: boolean read FSwapBytes write SetSwapBytes;

        // ,      
        procedure CheckTimeouts;
        constructor Create;
        function ReadDInputs(Adress: byte; Start: word; Quantity: word; out ErrorCode: byte; out TransID: word): rawbytestring;
        function ReadCoils(Adress: byte; Start: word; Quantity: word; out ErrorCode: byte; out TransID: word): rawbytestring;
        function ReadHoldingRegs(Adress: byte; Start: word; Quantity: word; out ErrorCode: byte;
            out TransID: word): rawbytestring;
        function ReadInputRegs(Adress: byte; Start: word; Quantity: word; out ErrorCode: byte;
            out TransID: word): rawbytestring;
        function WriteOneCoil(Adress: byte; Start: word; State: boolean; out ErrorCode: byte;
            out TransID: word): rawbytestring;
        function WriteOneReg(Adress: byte; Start: word; Value: word; out ErrorCode: byte;
            out TransID: word): rawbytestring;
        function WriteCoils(Adress: byte; Start: word; Quantity: word; aData: rawbytestring;
            out ErrorCode: byte; out TransID: word): rawbytestring;
        function WriteRegs(Adress: byte; Start: word; Quantity: word; aData: rawbytestring;
            out ErrorCode: byte; out TransID: word): rawbytestring;
        function WriteMaskReg(Adress: byte; Start: word; AndMask: word; OrMask: word; out ErrorCode: byte;
            out TransID: word): rawbytestring;
        function ReadWriteRegs(Adress: byte; StartRead: word; QuantityRead: word; StartWrite: word;
            QuantityWrite: word; aData: rawbytestring; out ErrorCode: byte; out TransID: word): rawbytestring;
        procedure ReadResponce(Resp: rawbytestring);
    end;

function ModbusTcpSlaveMessage(var PLCMems: TPLCStations; out ErrorCode: byte; InputString: rawbytestring): rawbytestring;

function MB_WordToStrBE(w: word): rawbytestring; inline;

function MB_WordToStrLE(w: word): rawbytestring; inline;

function MB_StrToWord(c1, c2: char): word; inline;


implementation

function ModbusTcpSlaveMessage(var PLCMems: TPLCStations; out ErrorCode: byte; InputString: rawbytestring): rawbytestring;
var
    TmpWord: word;
    s: ansistring;
    idTrans: word;
    RequestLen: word;
    adr: byte;
    func: byte;
    Quantity: word;
    Count: byte;
    StartAdr: word;
    tmpByte: byte;
    tmpStr: ansistring;
    i, j: integer;
    OutValue: word;
    AndMask: word;
    OrMask: word;
    WriteAdr: word;
    WriteQuan: word;
begin
    Result := '';
    ErrorCode := 0;
    s := InputString;

    if InputString = '' then
    begin
        ErrorCode := 255;
        exit;
    end;

    if length(InputString) < 9 then
    begin
        ErrorCode := 254;
        exit;
    end;

    idTrans := (Ord(s[1]) shl 8) or Ord(s[2]);
    RequestLen := (Ord(s[5]) shl 8) or Ord(s[6]);
    adr := Ord(s[7]);
    func := Ord(s[8]);

    if length(s) <> (RequestLen + 6) then
    begin
        ErrorCode := 253;
        Result := LeftStr(s, 4) + #00 + #03 + s[7] + char(func or $80) + #03;
        exit;
    end;

    if not (func in [01, 02, 03, 04, 05, 06, 15, 16, 22, 23]) then
    begin
        ErrorCode := 252;
        Result := LeftStr(s, 4) + #00 + #03 + s[7] + char(func or $80) + #01;
        exit;
    end;

    if (func in [01, 02, 03, 04, 05, 06]) and (length(s) <> 12) then
    begin
        ErrorCode := 251;
        Result := LeftStr(s, 4) + #00 + #03 + s[7] + char(func or $80) + #03;
        exit;
    end;

    if PLCMems[adr] = nil then
    begin
        ErrorCode := 230;
        Result := LeftStr(s, 4) + #00 + #03 + s[7] + char(func or $80) + #04;
        exit;
    end;

    {$Region Func_01}
    if func = 01 then
    begin

        StartAdr := (Ord(s[9]) shl 8) or Ord(s[10]);
        Quantity := (Ord(s[11]) shl 8) or Ord(s[12]);

        if Quantity > 2000 then
        begin
            ErrorCode := 250;
            Result := LeftStr(s, 4) + #00 + #03 + s[7] + char(func or $80) + #03;
            exit;
        end;

        if (StartAdr + Quantity) > $10000 then
        begin
            ErrorCode := 249;
            Result := LeftStr(s, 4) + #00 + #03 + s[7] + char(func or $80) + #02;
            exit;
        end;

        // request is correct
        tmpStr := '';
        for i := 0 to (Quantity div 8) - 1 do
        begin
            tmpByte := 0;
            for j := 0 to 7 do
            begin
                if PLCMems[adr]^.Coils[StartAdr + i * 8 + j] then
                    tmpByte := tmpByte or (1 shl j);
            end;
            tmpStr := tmpStr + char(tmpByte);
        end;
        if (Quantity mod 8) <> 0 then
        begin
            i := Quantity and $F8;
            tmpByte := 0;
	        for j := 0 to (Quantity mod 8) - 1 do
	        begin
	            if PLCMems[adr]^.Coils[StartAdr + i + j] then
	                tmpByte := tmpByte or (1 shl j);
	        end;
            tmpStr := tmpStr + char(tmpByte);
        end;

        // [idTran(2) + idProtocol(2)] + [MessageLenght(2)] + [adr(1)] + [func(1)] + [ByteCount(1)] + [Responce(N)]
        Result := LeftStr(s, 4) + char((Length(tmpStr) + 3) shr 8) + char((Length(tmpStr) + 3) and $FF) +
            s[7] + s[8] + char(Length(tmpStr)) + tmpStr;
        exit;
    end;
    {$Endregion Func_01}

    {$Region Func_02}
    if func = 02 then
    begin

        StartAdr := (Ord(s[9]) shl 8) or Ord(s[10]);
        Quantity := (Ord(s[11]) shl 8) or Ord(s[12]);

        if Quantity > 2000 then
        begin
            ErrorCode := 250;
            Result := LeftStr(s, 4) + #00 + #03 + s[7] + char(func or $80) + #03;
            exit;
        end;

        if (StartAdr + Quantity) > $10000 then
        begin
            ErrorCode := 249;
            Result := LeftStr(s, 4) + #00 + #03 + s[7] + char(func or $80) + #02;
            exit;
        end;

        // request is correct
        tmpStr := '';
        for i := 0 to (Quantity div 8) - 1 do
        begin
            tmpByte := 0;
            for j := 0 to 7 do
            begin
                if PLCMems[adr]^.DInputs[StartAdr + i * 8 + j] then
                    tmpByte := tmpByte or (1 shl j);
            end;
            tmpStr := tmpStr + char(tmpByte);
        end;
        if (Quantity mod 8) <> 0 then
        begin
            i := Quantity and $F8;
            tmpByte := 0;
	        for j := 0 to (Quantity mod 8) - 1 do
	        begin
	            if PLCMems[adr]^.DInputs[StartAdr + i + j] then
	                tmpByte := tmpByte or (1 shl j);
	        end;
            tmpStr := tmpStr + char(tmpByte);
        end;
        // [idTran(2) + idProtocol(2)] + [MessageLenght(2)] + [adr(1)] + [func(1)] + [ByteCount(1)] + [Responce(N)]
        Result := LeftStr(s, 4) + char((Length(tmpStr) + 3) shr 8) + char((Length(tmpStr) + 3) and $FF) +
            s[7] + s[8] + char(Length(tmpStr)) + tmpStr;
        exit;
    end;
    {$Endregion Func_02}

    {$Region Func_03}
    if func = 03 then
    begin

        StartAdr := (Ord(s[9]) shl 8) or Ord(s[10]);
        Quantity := (Ord(s[11]) shl 8) or Ord(s[12]);

        if Quantity > 125 then
        begin
            ErrorCode := 250;
            Result := LeftStr(s, 4) + #00 + #03 + s[7] + char(func or $80) + #03;
            exit;
        end;

        if (StartAdr + Quantity) > $10000 then
        begin
            ErrorCode := 249;
            Result := LeftStr(s, 4) + #00 + #03 + s[7] + char(func or $80) + #02;
            exit;
        end;

        // request is correct
        tmpStr := '';
        for i := 0 to Quantity - 1 do
            tmpStr := tmpStr + char(PLCMems[adr]^.HoldingReg[StartAdr + i] shr 8) +
                char(PLCMems[adr]^.HoldingReg[StartAdr + i] and $FF);
        // [idTran(2) + idProtocol(2)] + [MessageLenght(2)] + [adr(1)] + [func(1)] + [ByteCount(1)] + [Responce(N)]
        Result := LeftStr(s, 4) + char((Length(tmpStr) + 3) shr 8) + char((Length(tmpStr) + 3) and $FF) +
            s[7] + s[8] + char(Length(tmpStr)) + tmpStr;
        exit;
    end;
    {$Endregion Func_03}

    {$Region Func_04}
    if func = 04 then
    begin

        StartAdr := (Ord(s[9]) shl 8) or Ord(s[10]);
        Quantity := (Ord(s[11]) shl 8) or Ord(s[12]);

        if Quantity > 125 then
        begin
            ErrorCode := 250;
            Result := LeftStr(s, 4) + #00 + #03 + s[7] + char(func or $80) + #03;
            exit;
        end;

        if (StartAdr + Quantity) > $10000 then
        begin
            ErrorCode := 249;
            Result := LeftStr(s, 4) + #00 + #03 + s[7] + char(func or $80) + #02;
            exit;
        end;

        // request is correct
        tmpStr := '';
        for i := 0 to Quantity - 1 do
            tmpStr := tmpStr + char(PLCMems[adr]^.InputReg[StartAdr + i] shr 8) +
                char(PLCMems[adr]^.InputReg[StartAdr + i] and $FF);
        // [idTran(2) + idProtocol(2)] + [MessageLenght(2)] + [adr(1)] + [func(1)] + [ByteCount(1)] + [Responce(N)]
        Result := LeftStr(s, 4) + char((Length(tmpStr) + 3) shr 8) + char((Length(tmpStr) + 3) and $FF) +
            s[7] + s[8] + char(Length(tmpStr)) + tmpStr;
        exit;
    end;
    {$Endregion Func_04}

    {$Region Func_05}
    if func = 05 then
    begin

        StartAdr := (Ord(s[9]) shl 8) or Ord(s[10]);
        OutValue := (Ord(s[11]) shl 8) or Ord(s[12]);

        if (OutValue <> 0) and (OutValue <> $FF00) then
        begin
            ErrorCode := 250;
            Result := LeftStr(s, 4) + #00 + #03 + s[7] + char(func or $80) + #03;
            exit;
        end;

        // request is correct
        if OutValue = $FF00 then
            PLCMems[adr]^.Coils[StartAdr] := True
        else
            PLCMems[adr]^.Coils[StartAdr] := False;
        Result := s;
        exit;
    end;
    {$Endregion Func_05}

    {$Region Func_06}
    if func = 06 then
    begin

        StartAdr := (Ord(s[9]) shl 8) or Ord(s[10]);
        OutValue := (Ord(s[11]) shl 8) or Ord(s[12]);

        // request is correct
        PLCMems[adr]^.HoldingReg[StartAdr] := OutValue;
        Result := s;
        exit;
    end;
    {$Endregion Func_06}

    {$Region Func_15}
    if func = 15 then
    begin
        if length(s) < 14 then
        begin
            ErrorCode := 251;
            Result := LeftStr(s, 4) + #00 + #03 + s[7] + char(func or $80) + #03;
            exit;
        end;

        StartAdr := (Ord(s[9]) shl 8) or Ord(s[10]);
        Quantity := (Ord(s[11]) shl 8) or Ord(s[12]);
        Count := Ord(s[13]);

        if Quantity > 2000 then
        begin
            ErrorCode := 250;
            Result := LeftStr(s, 4) + #00 + #03 + s[7] + char(func or $80) + #03;
            exit;
        end;

        if (StartAdr + Quantity) > $10000 then
        begin
            ErrorCode := 249;
            Result := LeftStr(s, 4) + #00 + #03 + s[7] + char(func or $80) + #02;
            exit;
        end;

        if (Quantity mod 8) = 0 then
        begin
            if Count <> (Quantity div 8) then
            begin
                ErrorCode := 248;
                Result := LeftStr(s, 4) + #00 + #03 + s[7] + char(func or $80) + #03;
                exit;
            end;
        end
        else
        begin
            if Count <> ((Quantity div 8) + 1) then
            begin
                ErrorCode := 248;
                Result := LeftStr(s, 4) + #00 + #03 + s[7] + char(func or $80) + #03;
                exit;
            end;
        end;

        if Length(s) <> (13 + Count) then
        begin
            ErrorCode := 247;
            Result := LeftStr(s, 4) + #00 + #03 + s[7] + char(func or $80) + #03;
            exit;
        end;

        // request is correct
        for i := 0 to (Quantity div 8) - 1 do
        begin
            tmpByte := Ord(s[14 + i]);
            for j := 0 to 7 do
            begin
                if (tmpByte and 1) = 1 then
                    PLCMems[adr]^.Coils[StartAdr + i * 8 + j] := True
                else
                    PLCMems[adr]^.Coils[StartAdr + i * 8 + j] := False;
                tmpByte := tmpByte shr 1;
            end;
        end;
        // [idTran(2) + idProtocol(2)] + [MessageLenght(2)] + [adr(1)] + [func(1)] + [StartAdr(2)] + [Quantity(2)]
        Result := LeftStr(s, 4) + #00 + #06 + s[7] + s[8] + s[9] + s[10] + s[11] + s[12];
        exit;
    end;
    {$Endregion Func_15}

    {$Region Func_16}
    if func = 16 then
    begin
        if length(s) < 15 then
        begin
            ErrorCode := 251;
            Result := LeftStr(s, 4) + #00 + #03 + s[7] + char(func or $80) + #03;
            exit;
        end;

        StartAdr := (Ord(s[9]) shl 8) or Ord(s[10]);
        Quantity := (Ord(s[11]) shl 8) or Ord(s[12]);
        Count := Ord(s[13]);

        if Quantity > 123 then
        begin
            ErrorCode := 250;
            Result := LeftStr(s, 4) + #00 + #03 + s[7] + char(func or $80) + #03;
            exit;
        end;

        if (StartAdr + Quantity) > $10000 then
        begin
            ErrorCode := 249;
            Result := LeftStr(s, 4) + #00 + #03 + s[7] + char(func or $80) + #02;
            exit;
        end;

        if (Quantity * 2) <> Count then
        begin
            ErrorCode := 248;
            Result := LeftStr(s, 4) + #00 + #03 + s[7] + char(func or $80) + #03;
            exit;
        end;

        if Length(s) <> (13 + Count) then
        begin
            ErrorCode := 247;
            Result := LeftStr(s, 4) + #00 + #03 + s[7] + char(func or $80) + #03;
            exit;
        end;

        // request is correct
        for i := 0 to Quantity - 1 do
            PLCMems[adr]^.HoldingReg[StartAdr + i] := (Ord(s[14 + (i * 2)]) shl 8) or Ord(s[14 + (i * 2 + 1)]);
        // [idTran(2) + idProtocol(2)] + [MessageLenght(2)] + [adr(1)] + [func(1)] + [StartAdr(2)] + [Quantity(2)]
        Result := LeftStr(s, 4) + #00 + #06 + s[7] + s[8] + s[9] + s[10] + s[11] + s[12];
        exit;
    end;
    {$Endregion Func_16}

    {$Region Func_22}
    if func = 22 then
    begin
        if length(s) <> 14 then
        begin
            ErrorCode := 251;
            Result := LeftStr(s, 4) + #00 + #03 + s[7] + char(func or $80) + #03;
            exit;
        end;

        StartAdr := (Ord(s[9]) shl 8) or Ord(s[10]);
        AndMask := (Ord(s[11]) shl 8) or Ord(s[12]);
        OrMask := (Ord(s[13]) shl 8) or Ord(s[14]);

        // request is correct
        PLCMems[adr]^.HoldingReg[StartAdr] := (PLCMems[adr]^.HoldingReg[StartAdr] and AndMask) or
            (OrMask and (not AndMask));
        Result := s;
        exit;
    end;
    {$Endregion Func_22}

    {$Region Func_23}
    if func = 23 then
    begin
        if length(s) < 15 then
        begin
            ErrorCode := 251;
            Result := LeftStr(s, 4) + #00 + #03 + s[7] + char(func or $80) + #03;
            exit;
        end;

        StartAdr := (Ord(s[9]) shl 8) or Ord(s[10]);
        Quantity := (Ord(s[11]) shl 8) or Ord(s[12]);
        WriteAdr := (Ord(s[13]) shl 8) or Ord(s[14]);
        WriteQuan := (Ord(s[15]) shl 8) or Ord(s[16]);
        Count := Ord(s[17]);

        if Quantity > 125 then
        begin
            ErrorCode := 250;
            Result := LeftStr(s, 4) + #00 + #03 + s[7] + char(func or $80) + #03;
            exit;
        end;

        if (StartAdr + Quantity) > $10000 then
        begin
            ErrorCode := 249;
            Result := LeftStr(s, 4) + #00 + #03 + s[7] + char(func or $80) + #02;
            exit;
        end;

        if WriteQuan > 121 then
        begin
            ErrorCode := 246;
            Result := LeftStr(s, 4) + #00 + #03 + s[7] + char(func or $80) + #03;
            exit;
        end;

        if (WriteAdr + WriteQuan) > $10000 then
        begin
            ErrorCode := 245;
            Result := LeftStr(s, 4) + #00 + #03 + s[7] + char(func or $80) + #02;
            exit;
        end;

        if (WriteQuan * 2) <> Count then
        begin
            ErrorCode := 248;
            Result := LeftStr(s, 4) + #00 + #03 + s[7] + char(func or $80) + #03;
            exit;
        end;

        if Length(s) <> (17 + Count) then
        begin
            ErrorCode := 247;
            Result := LeftStr(s, 4) + #00 + #03 + s[7] + char(func or $80) + #03;
            exit;
        end;

        // request is correct
        tmpStr := '';
        for i := 0 to WriteQuan - 1 do
            PLCMems[adr]^.HoldingReg[StartAdr + i] := (Ord(s[18 + (i * 2)]) shl 8) or Ord(s[18 + (i * 2 + 1)]);
        for i := 0 to Quantity - 1 do
            tmpStr := tmpStr + char(PLCMems[adr]^.HoldingReg[StartAdr + i] shr 8) +
                char(PLCMems[adr]^.HoldingReg[StartAdr + i] and $FF);
        // [idTran(2) + idProtocol(2)] + [MessageLenght(2)] + [adr(1)] + [func(1)] + [ByteCount(1)] + [Responce(N)]
        Result := LeftStr(s, 4) + char((Length(tmpStr) + 3) shr 8) + char((Length(tmpStr) + 3) and $FF) +
            s[7] + s[8] + char(Length(tmpStr)) + tmpStr;
        exit;
    end;
    {$Endregion Func_23}

end;

function MB_WordToStrBE(w: word): rawbytestring;
begin
    Result := char(w shr 8) + char(w and $FF);
end;

function MB_WordToStrLE(w: word): rawbytestring;
begin
    Result := char(w and $FF) + char(w shr 8);
end;

function MB_StrToWord(c1, c2: char): word;
begin
    Result := ( ord(c1) shl 8 ) or ord(c2);
end;

{ TPascalModbasTCPMaster }

procedure TPascalModbasTCPMaster.SetTimeOut(AValue: word);
begin
    if FTimeOut = AValue then
        Exit;
    if AValue > 55000 then
        FTimeOut := 55000
    else if AValue < 20 then
        FTimeOut := 20
    else
        FTimeOut := AValue;
end;

procedure TPascalModbasTCPMaster.SetWriteOkProcedure(AValue: TModbusTCPWriteOk);
begin
    if FWriteOkProcedure = AValue then
        Exit;
    FWriteOkProcedure := AValue;
end;

//        TransID
procedure TPascalModbasTCPMaster.GetNextTransID;
var
    i,j,k: word;
begin

    i := NextTransID shr 11;            //  5  (0..31)
    j := (NextTransID and $7FF) shr 6;  //  5  (0..31)
    k := NextTransID and $3F;           //  6  (0..63)
    //    
    TransMap3[i] := true;
    TransMap2[i,j] := true;
    TransMap1[i,j,k] := true;

    //    
    if FActiveRequests < 65535 then
        Inc(FActiveRequests);
    //    TransID
    i := 1;
    Inc(NextTransID);
    while Transactions[NextTransID].Func <> 0 do
    begin
        Inc(NextTransID);
        Inc(i);
        if i >= 65535 then
            Break;
    end;
end;

procedure TPascalModbasTCPMaster.DeactivateTransaction(const RTransID: word);
var
	active: Boolean;
	n: word;
	k: word;
	j: word;
	i: word;
begin
    //    .  ID     .
    Transactions[RTransID].Func := 0;
    //  
    i := RTransID shr 11;
    j := (RTransID shr 6) and $1F;
    k := RTransID and $3F;
    //    
    if FActiveRequests > 0 then
        Dec(FActiveRequests);
    //         
    TransMap1[i,j,k] := false;
    //   .     
    active := false;
    //     
    for n := 0 to 63 do
        if TransMap1[i,j,n] then
            active := true;
    //    -     
    if not active then
        TransMap2[i,j] := false;

    //    
    active := false;
    //     
    for n := 0 to 31 do
        if TransMap2[i,n] then
            active := true;
    //    -     
    if not active then
        TransMap3[i] := false;
end;

procedure TPascalModbasTCPMaster.SetRequestTimeoutProcedure(AValue: TModbusTCPRequestTimeout);
begin
    if FRequestTimeoutProcedure = AValue then
        Exit;
    FRequestTimeoutProcedure := AValue;
end;

procedure TPascalModbasTCPMaster.SetSwapBytes(AValue: boolean);
begin
    if FSwapBytes = AValue then
        Exit;
    FSwapBytes := AValue;
end;

procedure TPascalModbasTCPMaster.SetRequestErrorProcedure(AValue: TModbusTCPRequestError);
begin
    if FRequestErrorProcedure = AValue then
        Exit;
    FRequestErrorProcedure := AValue;
end;

procedure TPascalModbasTCPMaster.SetReadBitsProcedure(AValue: TModbusTCPReadBits);
begin
    if FReadBitsProcedure = AValue then
        Exit;
    FReadBitsProcedure := AValue;
end;

procedure TPascalModbasTCPMaster.SetMaxActiveRequests(AValue: word);
begin
    if FMaxActiveRequests = AValue then
        Exit;
    FMaxActiveRequests := AValue;
end;

procedure TPascalModbasTCPMaster.SetReadWordsProcedure(AValue: TModbusTCPReadWords);
begin
    if FReadWordsProcedure = AValue then
        Exit;
    FReadWordsProcedure := AValue;
end;

procedure TPascalModbasTCPMaster.CheckTimeouts;
var
    ThisTime: word;
    i, j, k: integer;
    ActiveI, ActiveJ: boolean;
begin
    //         
    ThisTime := MilliSecondOfTheMinute(now);
    //  
    for i := 0 to 31 do
    begin
        //          
        if TransMap3[i] then
        begin
            // ,  ,       
            ActiveI := false;
            //    
            for j := 0 to 31 do
            begin
                //          
                if TransMap2[i,j] then
                begin
                    // ,  ,       
                    ActiveJ := false;
                    //   
                    for k := 0 to 63 do
                    begin
                        // ,     ( )
                        if TransMap1[i,j,k] then
                        begin
                            //  
                            if (ThisTime - Transactions[i * 32 * 64 + j * 64 + k].StartTime) >= FTimeOut then
                            begin
                                //    -     (  )
                                if assigned(FRequestTimeoutProcedure) then
                                    FRequestTimeoutProcedure(Transactions[i * 32 * 64 + j * 64 + k]);
                                //    
                                if FActiveRequests > 0 then
                                    Dec(FActiveRequests);
                                //         
                                TransMap1[i,j,k] := false;
                                //       
                                Transactions[i * 32 * 64 + j * 64 + k].Func := 0;
                            end
                            //        
                            else
                                // ,         
                                ActiveJ := true;
                        end;
                    end;
                    //           
                    if not ActiveJ then
                        //      
                        TransMap2[i,j] := false
                    else
                        //      -        
                        ActiveI := true;
                end;
            end;
            //          
            if not ActiveI then
                //    
                TransMap3[i] := false;
        end;
    end;

end;

constructor TPascalModbasTCPMaster.Create;
var
    i,j,k: word;
begin
    for i := 0 to $FFFF do
    begin
        Transactions[i].Adress := 0;
        Transactions[i].Func := 0;
        Transactions[i].StartReg := 0;
        Transactions[i].CountReg := 0;
        Transactions[i].WriteReg := 0;
        Transactions[i].WriteCount := 0;
        Transactions[i].StartTime := 0;
    end;
    for i := 0 to 31 do
    begin
        TransMap3[i] := false;
        for j := 0 to 31 do
        begin
            TransMap2[i,j] := false;
            for k := 0 to 63 do
                TransMap1[i,j,k] := false;
        end;
    end;
    FTimeout := 1000;
    FRequestTimeoutProcedure := nil;
    FRequestErrorProcedure := nil;
    FWriteOkProcedure := nil;
    FReadBitsProcedure := nil;
    FReadWordsProcedure := nil;
    NextTransID := 0;
    SwapBytes := False;
    FMaxActiveRequests := 10;
    FActiveRequests := 0;
end;

function TPascalModbasTCPMaster.ReadDInputs(Adress: byte; Start: word; Quantity: word; out ErrorCode: byte;
    out TransID: word): rawbytestring;
begin
    //       
    Result := '';
    ErrorCode := 0;
    //    
    if (Start + Quantity) > $100000 then
        ErrorCode := 255;
    //    
    if (Quantity > 2000) or (Quantity = 0) then
        ErrorCode := 254;
    //   
    if FActiveRequests >= FMaxActiveRequests then
        ErrorCode := 200;
    //   ,   
    TransID := NextTransID;
    //   ,    
    if ErrorCode <> 0 then
        exit;

    //  .  
    // TransID + ProtocolID + Length + Adress + Function + Start + Quantity
    Result := MB_WordToStrBE(NextTransID) + #00 + #00 + MB_WordToStrBE(6) + char(Adress) + char(02) +
        MB_WordToStrBE(Start) + MB_WordToStrBE(Quantity);

    //    
    Transactions[NextTransID].Adress := Adress;
    Transactions[NextTransID].Func := 2;
    Transactions[NextTransID].StartReg := Start;
    Transactions[NextTransID].CountReg := Quantity;
    Transactions[NextTransID].WriteReg := 0;
    Transactions[NextTransID].WriteCount := 0;
    Transactions[NextTransID].StartTime := MilliSecondOfTheMinute(now);

    //   ID
    GetNextTransID;

end;

function TPascalModbasTCPMaster.ReadCoils(Adress: byte; Start: word; Quantity: word; out ErrorCode: byte;
    out TransID: word): rawbytestring;
begin
    //       
    Result := '';
    ErrorCode := 0;
    //    
    if (Start + Quantity) > $100000 then
        ErrorCode := 255;
    //    
    if (Quantity > 2000) or (Quantity = 0) then
        ErrorCode := 254;
    //   
    if FActiveRequests >= FMaxActiveRequests then
        ErrorCode := 200;
    //   ,   
    TransID := NextTransID;
    //   ,    
    if ErrorCode <> 0 then
        exit;

    //  .  
    // TransID + ProtocolID + Length + Adress + Function + Start + Quantity
    Result := MB_WordToStrBE(NextTransID) + #00 + #00 + MB_WordToStrBE(6) + char(Adress) + char(01) +
        MB_WordToStrBE(Start) + MB_WordToStrBE(Quantity);

    //    
    Transactions[NextTransID].Adress := Adress;
    Transactions[NextTransID].Func := 1;
    Transactions[NextTransID].StartReg := Start;
    Transactions[NextTransID].CountReg := Quantity;
    Transactions[NextTransID].WriteReg := 0;
    Transactions[NextTransID].WriteCount := 0;
    Transactions[NextTransID].StartTime := MilliSecondOfTheMinute(now);

    //   ID
    GetNextTransID;
end;

function TPascalModbasTCPMaster.ReadHoldingRegs(Adress: byte; Start: word; Quantity: word;
    out ErrorCode: byte; out TransID: word): rawbytestring;
begin
    //       
    Result := '';
    ErrorCode := 0;
    //    
    if (Start + Quantity) > $100000 then
        ErrorCode := 255;
    //    
    if (Quantity > 125) or (Quantity = 0) then
        ErrorCode := 254;
    //   
    if FActiveRequests >= FMaxActiveRequests then
        ErrorCode := 200;
    //   ,   
    TransID := NextTransID;
    //   ,    
    if ErrorCode <> 0 then
        exit;

    //  .  
    // TransID + ProtocolID + Length + Adress + Function + Start + Quantity
    Result := MB_WordToStrBE(NextTransID) + #00 + #00 + MB_WordToStrBE(6) + char(Adress) + char(03) +
        MB_WordToStrBE(Start) + MB_WordToStrBE(Quantity);

    //    
    Transactions[NextTransID].Adress := Adress;
    Transactions[NextTransID].Func := 3;
    Transactions[NextTransID].StartReg := Start;
    Transactions[NextTransID].CountReg := Quantity;
    Transactions[NextTransID].WriteReg := 0;
    Transactions[NextTransID].WriteCount := 0;
    Transactions[NextTransID].StartTime := MilliSecondOfTheMinute(now);

    //   ID
    GetNextTransID;
end;

function TPascalModbasTCPMaster.ReadInputRegs(Adress: byte; Start: word; Quantity: word; out ErrorCode: byte;
    out TransID: word): rawbytestring;
begin
    //       
    Result := '';
    ErrorCode := 0;
    //    
    if (Start + Quantity) > $100000 then
        ErrorCode := 255;
    //    
    if (Quantity > 125) or (Quantity = 0) then
        ErrorCode := 254;
    //   
    if FActiveRequests >= FMaxActiveRequests then
        ErrorCode := 200;
    //   ,   
    TransID := NextTransID;
    //   ,    
    if ErrorCode <> 0 then
        exit;

    //  .  
    // TransID + ProtocolID + Length + Adress + Function + Start + Quantity
    Result := MB_WordToStrBE(NextTransID) + #00 + #00 + MB_WordToStrBE(6) + char(Adress) + char(04) +
        MB_WordToStrBE(Start) + MB_WordToStrBE(Quantity);

    //    
    Transactions[NextTransID].Adress := Adress;
    Transactions[NextTransID].Func := 4;
    Transactions[NextTransID].StartReg := Start;
    Transactions[NextTransID].CountReg := Quantity;
    Transactions[NextTransID].WriteReg := 0;
    Transactions[NextTransID].WriteCount := 0;
    Transactions[NextTransID].StartTime := MilliSecondOfTheMinute(now);

    //   ID
    GetNextTransID;
end;

function TPascalModbasTCPMaster.WriteOneCoil(Adress: byte; Start: word; State: boolean; out ErrorCode: byte;
    out TransID: word): rawbytestring;
var
    Value: word;
begin
    //       
    Result := '';
    ErrorCode := 0;
    //   
    if FActiveRequests >= FMaxActiveRequests then
        ErrorCode := 200;
    //   ,   
    TransID := NextTransID;
    //   ,    
    if ErrorCode <> 0 then
        exit;

    //  .  
    if State then
        Value := $FF00
    else
        Value := 0;
    // TransID + ProtocolID + Length + Adress + Function + Start + Quantity
    Result := MB_WordToStrBE(NextTransID) + #00 + #00 + MB_WordToStrBE(6) + char(Adress) + char(05) +
        MB_WordToStrBE(Start) + MB_WordToStrBE(Value);

    //    
    Transactions[NextTransID].Adress := Adress;
    Transactions[NextTransID].Func := 5;
    Transactions[NextTransID].StartReg := 0;
    Transactions[NextTransID].CountReg := 0;
    Transactions[NextTransID].WriteReg := Start;
    Transactions[NextTransID].WriteCount := Value;
    Transactions[NextTransID].StartTime := MilliSecondOfTheMinute(now);

    //   ID
    GetNextTransID;
end;

function TPascalModbasTCPMaster.WriteOneReg(Adress: byte; Start: word; Value: word; out ErrorCode: byte;
    out TransID: word): rawbytestring;
begin
    //       
    Result := '';
    ErrorCode := 0;
    //   
    if FActiveRequests >= FMaxActiveRequests then
        ErrorCode := 200;
    //   ,   
    TransID := NextTransID;
    //   ,    
    if ErrorCode <> 0 then
        exit;

    //  .  
    // TransID + ProtocolID + Length + Adress + Function + Start + Quantity
    Result := MB_WordToStrBE(NextTransID) + #00 + #00 + MB_WordToStrBE(6) + char(Adress) + char(06) +
        MB_WordToStrBE(Start) + MB_WordToStrBE(Value);

    //    
    Transactions[NextTransID].Adress := Adress;
    Transactions[NextTransID].Func := 6;
    Transactions[NextTransID].StartReg := 0;
    Transactions[NextTransID].CountReg := 0;
    Transactions[NextTransID].WriteReg := Start;
    Transactions[NextTransID].WriteCount := Value;
    Transactions[NextTransID].StartTime := MilliSecondOfTheMinute(now);

    //   ID
    GetNextTransID;
end;

function TPascalModbasTCPMaster.WriteCoils(Adress: byte; Start: word; Quantity: word; aData: rawbytestring;
    out ErrorCode: byte; out TransID: word): rawbytestring;
var
    TmpStr: rawbytestring; //    
    TmpByte: byte;
    i, j: integer;
begin
    //       
    Result := '';
    ErrorCode := 0;
    //    
    if (Start + Quantity) > $100000 then
        ErrorCode := 255;
    //    
    if (Quantity > 1968) or (Quantity = 0) then
        ErrorCode := 254;
    //    
    if length(aData) <> Quantity then
        ErrorCode := 253;
    //   
    if FActiveRequests >= FMaxActiveRequests then
        ErrorCode := 200;
    //   ,   
    TransID := NextTransID;
    //   ,    
    if ErrorCode <> 0 then
        exit;

    //  .  
    //     
    TmpStr := '';
    //   
    for i := 0 to (Quantity div 8) - 1 do
    begin
        TmpByte := 0;
        //  
        for j := 1 to 8 do
        begin
            // .       
            TmpByte := TmpByte shr 1;
            //   "  00.  - 
            if aData[i * 8 + j] <> #00 then
                TmpByte := TmpByte or $80;
        end;
        //  
        TmpStr := TmpStr + char(TmpByte);
    end;
    //      (   )
    j := length(TmpStr) * 8;
    //    
    if (Quantity mod 8) <> 0 then
    begin
        TmpByte := 0;
        for i := 0 to (Quantity mod 8) - 1 do
        begin
            //   
            if aData[j + i + 1] <> #00 then
                //     
                TmpByte := TmpByte or (1 shl i);
        end;
        //  
        TmpStr := TmpStr + char(TmpByte);
    end;
    // TransID + ProtocolID + Length + Adress + Function + Start + Quantity
    Result := MB_WordToStrBE(NextTransID) + #00 + #00 + MB_WordToStrBE(7 + length(TmpStr)) + char(Adress) + char(15) +
        MB_WordToStrBE(Start) + MB_WordToStrBE(Quantity) + char(length(TmpStr)) + TmpStr;

    //    
    Transactions[NextTransID].Adress := Adress;
    Transactions[NextTransID].Func := 15;
    Transactions[NextTransID].StartReg := 0;
    Transactions[NextTransID].CountReg := 0;
    Transactions[NextTransID].WriteReg := Start;
    Transactions[NextTransID].WriteCount := Quantity;
    Transactions[NextTransID].StartTime := MilliSecondOfTheMinute(now);

    //   ID
    GetNextTransID;
end;

function TPascalModbasTCPMaster.WriteRegs(Adress: byte; Start: word; Quantity: word; aData: rawbytestring;
    out ErrorCode: byte; out TransID: word): rawbytestring;
begin
    //       
    Result := '';
    ErrorCode := 0;
    //    
    if (Start + Quantity) > $100000 then
        ErrorCode := 255;
    //    
    if (Quantity > 123) or (Quantity = 0) then
        ErrorCode := 254;
    //    
    if length(aData) <> (Quantity * 2) then
        ErrorCode := 253;
    //   
    if FActiveRequests >= FMaxActiveRequests then
        ErrorCode := 200;
    //   ,   
    TransID := NextTransID;
    //   ,    
    if ErrorCode <> 0 then
        exit;

    //  .  
    // TransID + ProtocolID + Length + Adress + Function + Start + Quantity
    Result := MB_WordToStrBE(NextTransID) + #00 + #00 + MB_WordToStrBE(7 + length(aData)) + char(Adress) + char(16) +
        MB_WordToStrBE(Start) + MB_WordToStrBE(Quantity) + char(length(aData)) + aData;

    //    
    Transactions[NextTransID].Adress := Adress;
    Transactions[NextTransID].Func := 16;
    Transactions[NextTransID].StartReg := 0;
    Transactions[NextTransID].CountReg := 0;
    Transactions[NextTransID].WriteReg := Start;
    Transactions[NextTransID].WriteCount := Quantity;
    Transactions[NextTransID].StartTime := MilliSecondOfTheMinute(now);

    //   ID
    GetNextTransID;
end;

function TPascalModbasTCPMaster.WriteMaskReg(Adress: byte; Start: word; AndMask: word; OrMask: word;
    out ErrorCode: byte; out TransID: word): rawbytestring;
begin
    //       
    Result := '';
    ErrorCode := 0;
    //   
    if FActiveRequests >= FMaxActiveRequests then
        ErrorCode := 200;
    //   ,   
    TransID := NextTransID;
    //   ,    
    if ErrorCode <> 0 then
        exit;

    //  .  
    // TransID + ProtocolID + Length + Adress + Function + Start + Quantity
    Result := MB_WordToStrBE(NextTransID) + #00 + #00 + MB_WordToStrBE(8) + char(Adress) + char(22) +
        MB_WordToStrBE(Start) + MB_WordToStrBE(AndMask) + MB_WordToStrBE(OrMask);

    //    
    Transactions[NextTransID].Adress := Adress;
    Transactions[NextTransID].Func := 22;
    Transactions[NextTransID].StartReg := AndMask;
    Transactions[NextTransID].CountReg := 0;
    Transactions[NextTransID].WriteReg := Start;
    Transactions[NextTransID].WriteCount := OrMask;
    Transactions[NextTransID].StartTime := MilliSecondOfTheMinute(now);

    //   ID
    GetNextTransID;
end;

function TPascalModbasTCPMaster.ReadWriteRegs(Adress: byte; StartRead: word; QuantityRead: word;
    StartWrite: word; QuantityWrite: word; aData: rawbytestring; out ErrorCode: byte; out TransID: word): rawbytestring;
begin
    //       
    Result := '';
    ErrorCode := 0;
    //    
    if (StartWrite + QuantityWrite) > $100000 then
        ErrorCode := 255;
    if (StartRead + QuantityRead) > $100000 then
        ErrorCode := 255;
    //    
    if (QuantityWrite > 121) or (QuantityWrite = 0) or (QuantityRead > 125) or (QuantityRead = 0) then
        ErrorCode := 254;
    //    
    if length(aData) <> (QuantityWrite * 2) then
        ErrorCode := 253;
    //   
    if FActiveRequests >= FMaxActiveRequests then
        ErrorCode := 200;
    //   ,   
    TransID := NextTransID;
    //   ,    
    if ErrorCode <> 0 then
        exit;

    //  .  
    // TransID + ProtocolID + Length + Adress + Function + Start + Quantity
    Result := MB_WordToStrBE(NextTransID) + #00 + #00 + MB_WordToStrBE(11 + length(aData)) + char(Adress) + char(23) +
        MB_WordToStrBE(StartRead) + MB_WordToStrBE(QuantityRead) + MB_WordToStrBE(StartWrite) + MB_WordToStrBE(QuantityWrite) +
        char(length(aData)) + aData;

    //    
    Transactions[NextTransID].Adress := Adress;
    Transactions[NextTransID].Func := 23;
    Transactions[NextTransID].StartReg := StartRead;
    Transactions[NextTransID].CountReg := QuantityRead;
    Transactions[NextTransID].WriteReg := StartWrite;
    Transactions[NextTransID].WriteCount := QuantityWrite;
    Transactions[NextTransID].StartTime := MilliSecondOfTheMinute(now);

    //   ID
    GetNextTransID;
end;

procedure TPascalModbasTCPMaster.ReadResponce(Resp: rawbytestring);
var
    RTransID: word;
    RLength: word;
    RAdr: byte;
    RFunc: byte;
    RStartReg: word;
    RQuantity: word;
    ROrMask: word;
    RAndMask: word;
    RCount: byte;
    MBTrans: TModbusTransaction;
    n,l: word;
    DataString: rawbytestring;
    TmpByte: byte;
begin
    if length(Resp) < 2 then
    begin
        MBTrans.Adress := 0;
        MBTrans.Func := 0;
        MBTrans.CountReg := 0;
        MBTrans.WriteCount := 0;
        //   
        if Assigned(FRequestErrorProcedure) then FRequestErrorProcedure(MBTrans, 255, Resp);
        exit;
    end;
    RTransID := MB_StrToWord(Resp[1],Resp[2]);


    //     -      
    if length(Resp) < 9 then
    begin
        //   
        if Assigned(FRequestErrorProcedure) then FRequestErrorProcedure(Transactions[RTransID], 254, Resp);
        //  
		DeactivateTransaction(RTransID);
        //       . .
        exit;
    end;

    //   ,    9 

    RLength := MB_StrToWord(Resp[5],Resp[6]);
    RAdr := ord(Resp[7]);
    RFunc := ord(Resp[8]);

    //   (   )   
    if Transactions[RTransID].Func <> (RFunc and $7F) then
    begin
        if Assigned(FRequestErrorProcedure) then FRequestErrorProcedure(Transactions[RTransID], 253, Resp);
		DeactivateTransaction(RTransID);
        exit;
    end;

    //     
    if (RFunc and $80) <> 0 then
    begin
        //   .    ,   
        if Assigned(FRequestErrorProcedure) then FRequestErrorProcedure(Transactions[RTransID], 252, Resp);
		DeactivateTransaction(RTransID);
        exit;
    end;

    //     . ,    "  ".   .
    if not (RFunc in [01, 02, 03, 04, 05, 06, 15, 16, 22, 23]) then
    begin
        if Assigned(FRequestErrorProcedure) then FRequestErrorProcedure(Transactions[RTransID], 251, Resp);
		DeactivateTransaction(RTransID);
        exit;
    end;

    //      (  0)
    if (Transactions[RTransID].Adress <> 0) and (Transactions[RTransID].Adress <> RAdr) then
    begin
        if Assigned(FRequestErrorProcedure) then FRequestErrorProcedure(Transactions[RTransID], 250, Resp);
		DeactivateTransaction(RTransID);
        exit;
    end;

    //     
    if (RLength + 6) <> Length(Resp) then
    begin
        if Assigned(FRequestErrorProcedure) then FRequestErrorProcedure(Transactions[RTransID], 249, Resp);
		DeactivateTransaction(RTransID);
        exit;
    end;

    //     

    {$Region RFunc_01_02}
    //   01  02  
    if (RFunc = 01) or (RFunc = 02) then
    begin
        RCount := ord(Resp[9]);

        // :  
        if RLength <> (RCount + 3) then
        begin
            if Assigned(FRequestErrorProcedure) then FRequestErrorProcedure(Transactions[RTransID], 248, Resp);
			DeactivateTransaction(RTransID);
	        exit;
        end;

        // :   0 
        if RCount = 0 then
        begin
            if Assigned(FRequestErrorProcedure) then FRequestErrorProcedure(Transactions[RTransID], 247, Resp);
			DeactivateTransaction(RTransID);
	        exit;
        end;

        // :    
        if (Transactions[RTransID].CountReg mod 8) <> 0 then
	        if RCount <> (Transactions[RTransID].CountReg div 8 + 1) then
	        begin
	            if Assigned(FRequestErrorProcedure) then FRequestErrorProcedure(Transactions[RTransID], 246, Resp);
				DeactivateTransaction(RTransID);
		        exit;
	        end;
        if (Transactions[RTransID].CountReg mod 8) = 0 then
	        if RCount <> (Transactions[RTransID].CountReg div 8) then
	        begin
	            if Assigned(FRequestErrorProcedure) then FRequestErrorProcedure(Transactions[RTransID], 246, Resp);
				DeactivateTransaction(RTransID);
		        exit;
	        end;

        //  
        if Assigned(FReadBitsProcedure) then
        begin
            //       
            DataString := '';
            //    
            TmpByte := 0;
            //   
            l := 0;
            //   ,  .    -  
            for n := 0 to Transactions[RTransID].CountReg - 1 do
            begin
                //   
                TmpByte := ord( Resp[10 + (n div 8)] );
                //     
                if ( TmpByte and (1 shl l) ) <> 0 then
                    //  
                    DataString := DataString + '1'
                else
                    DataString := DataString + #00;
                //  
                inc(l);
                //   
                if l > 7 then
                    l := 0;
            end;

            //  
            FReadBitsProcedure(Transactions[RTransID], DataString);
        end;

        DeactivateTransaction(RTransID);
        exit;
    end;
    {$endregion RFunc_01_02}

    {$Region RFunc_03_04_23}
    //   03, 04  23  
    if (RFunc = 03) or (RFunc = 04) or (RFunc = 23) then
    begin
        RCount := ord(Resp[9]);

        // :  
        if RLength <> (RCount + 3) then
        begin
            if Assigned(FRequestErrorProcedure) then FRequestErrorProcedure(Transactions[RTransID], 248, Resp);
			DeactivateTransaction(RTransID);
	        exit;
        end;

        // :   0 
        if RCount = 0 then
        begin
            if Assigned(FRequestErrorProcedure) then FRequestErrorProcedure(Transactions[RTransID], 247, Resp);
			DeactivateTransaction(RTransID);
	        exit;
        end;

        // :    
        if RCount <> (Transactions[RTransID].CountReg * 2) then
        begin
            if Assigned(FRequestErrorProcedure) then FRequestErrorProcedure(Transactions[RTransID], 246, Resp);
			DeactivateTransaction(RTransID);
	        exit;
        end;

        //  
        if Assigned(FReadWordsProcedure) then
        begin
            //   
            DataString := RightStr(Resp, RCount);

            //  
            FReadWordsProcedure(Transactions[RTransID], DataString);
        end;

        DeactivateTransaction(RTransID);
        exit;
    end;
    {$endregion RFunc_03_04_23}

    {$Region RFunc_05_06_15_16}
    //   05, 06, 15  16  
    if (RFunc = 05) or (RFunc = 06) or (RFunc = 15) or (RFunc = 16) then
    begin
        // :  
        if RLength <> 6 then
        begin
            if Assigned(FRequestErrorProcedure) then FRequestErrorProcedure(Transactions[RTransID], 248, Resp);
			DeactivateTransaction(RTransID);
	        exit;
        end;

        RStartReg := MB_StrToWord(Resp[9], Resp[10]);
        RQuantity := MB_StrToWord(Resp[11], Resp[12]);

        // :   
        if (RFunc = 05) and (RQuantity <> 0) and (RQuantity <> $FF00) then
        begin
            if Assigned(FRequestErrorProcedure) then FRequestErrorProcedure(Transactions[RTransID], 245, Resp);
			DeactivateTransaction(RTransID);
	        exit;
        end;

        // :   0 
        if (RQuantity = 0) and (RFunc <> 05) then
        begin
            if Assigned(FRequestErrorProcedure) then FRequestErrorProcedure(Transactions[RTransID], 247, Resp);
			DeactivateTransaction(RTransID);
	        exit;
        end;

        // :   
        if RStartReg <> (Transactions[RTransID].WriteReg) then
        begin
            if Assigned(FRequestErrorProcedure) then FRequestErrorProcedure(Transactions[RTransID], 244, Resp);
			DeactivateTransaction(RTransID);
	        exit;
		end;

        // :    
        if RQuantity <> Transactions[RTransID].WriteCount then
        begin
            if Assigned(FRequestErrorProcedure) then FRequestErrorProcedure(Transactions[RTransID], 246, Resp);
			DeactivateTransaction(RTransID);
	        exit;
        end;

        //  
        if Assigned(FWriteOkProcedure) then
        begin
            //  
            FWriteOkProcedure(Transactions[RTransID]);
        end;

        DeactivateTransaction(RTransID);
        exit;
    end;
    {$endregion RFunc_05_06_15_16}

    {$Region RFunc_22}
    //   22  
    if (RFunc = 22) then
    begin
        // :  
        if RLength <> 8 then
        begin
            if Assigned(FRequestErrorProcedure) then FRequestErrorProcedure(Transactions[RTransID], 248, Resp);
			DeactivateTransaction(RTransID);
	        exit;
        end;

        RStartReg := MB_StrToWord(Resp[9], Resp[10]);
        RAndMask := MB_StrToWord(Resp[11], Resp[12]);
        ROrMask  := MB_StrToWord(Resp[13], Resp[14]);

        // :   
        if RStartReg <> (Transactions[RTransID].WriteReg) then
        begin
            if Assigned(FRequestErrorProcedure) then FRequestErrorProcedure(Transactions[RTransID], 244, Resp);
			DeactivateTransaction(RTransID);
	        exit;
		end;

        // :   
        if (RAndMask <> Transactions[RTransID].StartReg) or (ROrMask <> Transactions[RTransID].WriteCount) then
        begin
            if Assigned(FRequestErrorProcedure) then FRequestErrorProcedure(Transactions[RTransID], 243, Resp);
			DeactivateTransaction(RTransID);
	        exit;
        end;

        //  
        if Assigned(FWriteOkProcedure) then
        begin
            //  
            FWriteOkProcedure(Transactions[RTransID]);
        end;

        DeactivateTransaction(RTransID);
        exit;
    end;
    {$endregion RFunc_22}


end;


end.
